@@ -1,6 +1,5 @@
 # coding: US-ASCII
 require 'test/unit'
-require_relative 'envutil'
 
 class TestM17N < Test::Unit::TestCase
   def assert_encoding(encname, actual, message=nil)
@@ -261,6 +260,14 @@ class TestM17N < Test::Unit::TestCase
     end
   end
 
+  def test_utf_without_bom_asciionly
+    bug10598 = '[ruby-core:66835] [Bug #10598]'
+    encs = [Encoding::UTF_16, Encoding::UTF_32].find_all {|enc|
+      "abcd".force_encoding(enc).ascii_only?
+    }
+    assert_empty(encs, bug10598)
+  end
+
   def test_object_utf16_32_inspect
     EnvUtil.suppress_warning do
       begin
@@ -480,6 +487,9 @@ class TestM17N < Test::Unit::TestCase
     assert_regexp_fixed_ascii8bit(eval(a(%{/\xc2\xa1/n})))
     assert_regexp_fixed_ascii8bit(eval(a(%q{/\xc2\xa1/})))
 
+    s = '\xc2\xa1'
+    assert_regexp_fixed_ascii8bit(/#{s}/)
+
     assert_raise(SyntaxError) { eval("/\xa1\xa1/n".force_encoding("euc-jp")) }
 
     [/\xc2\xa1/n, eval(a(%{/\xc2\xa1/})), eval(a(%{/\xc2\xa1/n}))].each {|r|
@@ -1056,10 +1066,6 @@ class TestM17N < Test::Unit::TestCase
     assert_equal(false, e("\xa1\xa2\xa3\xa4").include?(e("\xa3")))
     s = e("\xa3\xb0\xa3\xb1\xa3\xb2\xa3\xb3\xa3\xb4")
     assert_equal(false, s.include?(e("\xb0\xa3")))
-    bug11488 = '[ruby-core:70592] [Bug #11488]'
-    each_encoding("abcdef", "def") do |str, substr|
-      assert_equal(true, str.include?(substr), bug11488)
-    end
   end
 
   def test_index
@@ -1069,10 +1075,6 @@ class TestM17N < Test::Unit::TestCase
     assert_nil(e("\xa1\xa2\xa3\xa4").rindex(e("\xa3")))
     s = e("\xa3\xb0\xa3\xb1\xa3\xb2\xa3\xb3\xa3\xb4")
     assert_raise(Encoding::CompatibilityError){s.rindex(a("\xb1\xa3"))}
-    bug11488 = '[ruby-core:70592] [Bug #11488]'
-    each_encoding("abcdef", "def") do |str, substr|
-      assert_equal(3, str.index(substr), bug11488)
-    end
   end
 
   def test_next
@@ -1140,12 +1142,7 @@ class TestM17N < Test::Unit::TestCase
   end
 
   def test_reverse
-    bug11387 = '[ruby-dev:49189] [Bug #11387]'
-    s1 = u("abcdefghij\xf0")
-    s2 = s1.reverse
-    assert_not_predicate(s1, :valid_encoding?, bug11387)
-    assert_equal(u("\xf0jihgfedcba"), s2)
-    assert_not_predicate(s2, :valid_encoding?, bug11387)
+    assert_equal(u("\xf0jihgfedcba"), u("abcdefghij\xf0").reverse)
   end
 
   def test_reverse_bang
@@ -1234,9 +1231,6 @@ class TestM17N < Test::Unit::TestCase
     each_encoding("abc,def", ",", "abc", "def") do |str, sep, *expected|
       assert_equal(expected, str.split(sep, -1))
     end
-    each_encoding("abc\0def", "\0", "abc", "def") do |str, sep, *expected|
-      assert_equal(expected, str.split(sep, -1))
-    end
   end
 
   def test_nonascii_method_name
